##################################################################
#
# (C) Copyright 2002 Kapil Thangavelu <kvthan@wm.edu>
# All Rights Reserved
#
# This file is part of CMFDeployment.
#
# CMFDeployment is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# CMFDeployment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

import re
import string
import unittest
from Acquisition import aq_base
from urlparse import urlparse

from Log import LogFactory
from BTrees.OOBTree import OOBTree

log = LogFactory('URIResolver')

class ResolutionError(Exception): pass

class URIResolver:
    """

    Creates a database of urls such that it can replace
    urls in rendered content to reflect new locations.

    uri database maps
    
     source_absolute_relative -> target_absolute_relative
    
     - target_absolute_relative = target_path + content_path_relative_to_mount_point
     - source_absolute_relative = content.absolute_relative(1)

    replaced url types

      all replaced urls are found with the url_regex below. it will find href and src
      urls. 

      - absolute - http://
      - relative - ../foo, bar/baz
      - absolute relative /foo/bar

    instances of this class need to be persisted for
    incremental deployments

    XXX

    replaced url types

      css2 import statements should get considered for replacment
    
      some additonal url types that should be considered

       - absolute url with relative segements http://foo/bar/../zebra.html
       - zpublisher marshal urls args (:method, :list) etc
    
    """
 
    def __init__(self,
                 id='',
                 source_host='http://localhost',
                 mount_path = '/',
                 target_path='/'):
        
        self.id = id
        self.source_host = source_host
        self.target_path = target_path
        self.mount_path = mount_path
        self.mlen = len(mount_path)
        self.uris = OOBTree()

    def addResource(self, descriptor):
        
        content = descriptor.getContent()

        # relative url
        # skins are really hard to remap, because
        # their referenced anywhere via acquisition:-(
        # still conditional logic here is generally
        # useful for structure remapping.
        
        relative_url = descriptor.getSourcePath() or content.absolute_url(relative=1)
        content_path = descriptor.getContentPath()

        if content_path is None:
            # minus last path segemnt
            url_context  = relative_url[:relative_url.rfind('/')] # find the path segment

            # content path relative to mount point
            # XXX this is bad, we're comparing physical to url paths
            assert url_context.find(self.mount_path) == 0
            
            content_path = url_context[self.mlen:]

        content_path = clstrip('/'.join( (self.target_path,
                                          content_path,
                                          descriptor.getFileName() ) ),
                               '/'
                               )
                                  
        self.uris['/'+relative_url]=content_path
                                 

    def resolve(self, descriptor):

        """
        regex to find all uris.
        classify as either
          - content relative - ../foo
          - absolute relative - /portal/content
          - absolute - http://foobz:8080/baz/blurbie
          
        relative uris, convert to absolute uris

        XXX for folder types, we replace to the view  not the
        folder path, it be nice to go to the path.

        # XXX
        # if we have a link reference to
        # non deployable content
        # we need some sort of policy decision
        """

        if descriptor.isBinary() or descriptor.isGhost():
            return

        r = descriptor.getRendered()
        urls = rlensort(filter(None, uri_regex.findall(r)))

        count = 0

        for u in urls:
            
            if not u:
                continue

            elif u.startswith('http'):
                # absolute

                # extern check
                if not u.find(self.source_host) >= 0:
                    #nu = None # continue # external ref
                    continue
                    
                # within mount check
                if u.count(self.mount_path) == 0:
                    nu = None

                # convert to absolute_relative and lookup
                nu = self.uris.get( urlparse(u)[2], None)
                
            elif u.startswith('/'):
                # absolute relative
                nu = self.uris.get('%s/%s'%(self.source_host,u), None)
                
            elif u.startswith('..') or u.startswith('.'):
                # content relative
                content = descriptor.getContent()
                content_folderish_p = getattr(aq_base(content), 'isObjectMangaer', 0)
                
                rnu = resolve_relative(content.absolute_url(1),
                                       u,
                                       content_folderish_p)
                
                nu = self.uris.get(rnu, None)
                
            elif u.startswith('#'):
                continue
            elif u.startswith('mailto'):
                continue
            elif u.startswith('ftp'):
                continue
            else:
                nu = None

            if nu is None:
                log.debug('unknown url (%s)'%u)
                nu = 'deployment_linking_error'
                #continue

            #log.debug('replace (%s) -> (%s)'%(u,nu) )
            r = r.replace(u, nu)
            count += 1
            
        log.debug('resolved %d references'%count)
        
        descriptor.setRendered(r)


def clstrip(string, character):
    # specialized (c)haracter (l)eft strip, leaves one occurence
    idx = 0
    while 1:
        if string[idx]==character:
            idx += 1
        else:
            break
    if idx > 1:
        return string[idx-1:]
    return string

def rlensort(lst):
    # sort a list of elements by length.
    lst2 = map(lambda x: (len(x), x), lst)
    lst2.sort()
    lst2.reverse()
    return [x[1] for x in lst2]

def resolve_relative(content_url, relative_url, content_folderish_p=0):

    # (addressing scheme, network location, path, query, fragment identifier)
    info = urlparse(content_url)
    content_path = info[2]
    
    path_steps = filter(None, content_path.split('/'))
    steps = filter(None, relative_url.split('/'))

    # if the first step is .. then we need to consider
    # whether the content itself is folderish or not
    # if its not folderish we remove two path steps
    # if its one.
    if steps[0] == '..':
        steps.pop(0)
        move = 1 + (not content_folderish_p)
        path_steps = path_steps[:-move]

    for s in steps:
        if s == '..':
            path_steps.pop(-1)
            continue            
        if s == '.':
            continue
        path_steps.append(s)

    path_steps.insert(0,'')
    resolved = '/'.join(path_steps)
    
    log.debug('relative (%s)->(%s)'%(relative_url, resolved))
    return resolved


uri_regex = re.compile('''(?:href=|src=|@import)\s*["']\s*(.*?)["']''')

#uri_regex = re.compile('''(?:href|src)=["']\s*(.*?)["']''')

class RelativeResolveTests(unittest.TestCase):

    def test1(self):
        a = 'http://www.example.com/fooz/ball.html'
        b = '../zebra/'
        expected = '/zebra'        
        db = {'/zebra':'/zebra'}

        import pdb ; pdb.set_trace()
        self.assertEqual(
            resolve_relative(a,b,db),
            expected,
            " bad resolve "
            )

    def test2(self):
        a = 'http://naz.com/fooz/barn/cat.html'
        b = './dog'
        
        
        
    
if __name__ == '__main__':
    unittest.main()
